# Copyright 2022 Mark Malek
# See LICENSE file for full license terms. 
""" Common named colors """

import pygame
BLACK = pygame.Color(0,0,0)
WHITE = pygame.Color(0xff,0xff,0xff)
GREEN = pygame.Color(0,255,0)
BLUE = pygame.Color(0,0,255)
RED = pygame.Color(255,0,0)
YELLOW = pygame.Color(255,255,0)
ORANGE = pygame.Color(255,128,0)
PURPLE = pygame.Color(255,0,255)